@@ -1,10 +1,8 @@
 # encoding: US-ASCII
 
-require 'rubygems'
-require 'minitest/autorun'
-require 'ruby_lexer'
-require 'ruby18_parser'
-require 'ruby20_parser'
+require "minitest/autorun"
+require "ruby_lexer"
+require "ruby_parser"
 
 class TestRubyLexer < Minitest::Test
   attr_accessor :processor, :lex, :parser_class, :lex_state
@@ -14,7 +12,7 @@ class TestRubyLexer < Minitest::Test
 
   def setup
     self.lex_state = :expr_beg
-    setup_lexer_class Ruby20Parser
+    setup_lexer_class RubyParser.latest.class
   end
 
   def setup_lexer input, exp_sexp = nil
@@ -53,6 +51,10 @@ class TestRubyLexer < Minitest::Test
     assert_lex(input, exp_sexp, *args, &block)
   end
 
+  def ruby18
+    RubyParser::V18 === lexer.parser
+  end
+
   def refute_lex input, *args # TODO: re-sort
     args = args.each_slice(2).map { |a, b| [a, b, nil, nil, nil] }.flatten
 
@@ -89,6 +91,8 @@ class TestRubyLexer < Minitest::Test
     case value
     when Float then
       assert_in_epsilon value, act_value, 0.001, msg
+    when NilClass then
+      assert_nil act_value, msg
     else
       assert_equal value, act_value,       msg
     end
@@ -246,13 +250,13 @@ class TestRubyLexer < Minitest::Test
   end
 
   def test_yylex_and_dot
-    setup_lexer_class Ruby23Parser
+    setup_lexer_class RubyParser::V23
 
     assert_lex3("&.", nil, :tLONELY, "&.", :expr_dot)
   end
 
   def test_yylex_and_dot_call
-    setup_lexer_class Ruby23Parser
+    setup_lexer_class RubyParser::V23
 
     assert_lex3("x&.y", nil,
                 :tIDENTIFIER, "x", :expr_cmdarg,
@@ -260,6 +264,15 @@ class TestRubyLexer < Minitest::Test
                 :tIDENTIFIER, "y")
   end
 
+  def test_yylex_and_dot_call_newline
+    setup_lexer_class Ruby23Parser
+
+    assert_lex3("x\n&.y", nil,
+                :tIDENTIFIER, "x", :expr_cmdarg,
+                :tLONELY, "&.", :expr_dot,
+                :tIDENTIFIER, "y")
+  end
+
   def test_yylex_and_arg
     self.lex_state = :expr_arg
 
@@ -292,7 +305,7 @@ class TestRubyLexer < Minitest::Test
   end
 
   def test_yylex_label__18
-    setup_lexer_class Ruby18Parser
+    setup_lexer_class RubyParser::V18
 
     assert_lex3("{a:",
                 nil,
@@ -302,7 +315,7 @@ class TestRubyLexer < Minitest::Test
   end
 
   def test_yylex_label_in_params__18
-    setup_lexer_class Ruby18Parser
+    setup_lexer_class RubyParser::V18
 
     assert_lex3("foo(a:",
                 nil,
@@ -313,7 +326,7 @@ class TestRubyLexer < Minitest::Test
   end
 
   def test_yylex_label__19
-    setup_lexer_class Ruby19Parser
+    setup_lexer_class RubyParser::V19
 
     assert_lex3("{a:",
                 nil,
@@ -322,7 +335,7 @@ class TestRubyLexer < Minitest::Test
   end
 
   def test_yylex_label_in_params__19
-    setup_lexer_class Ruby19Parser
+    setup_lexer_class RubyParser::V19
 
     assert_lex3("foo(a:",
                 nil,
@@ -381,7 +394,9 @@ class TestRubyLexer < Minitest::Test
     refute_lexeme
   end
 
-  def test_yylex_not_at_defn
+  def test_yylex_not_at_defn__20
+    setup_lexer_class RubyParser::V20
+
     assert_lex("def +@; end",
                s(:defn, :+@, s(:args), s(:nil)),
 
@@ -465,7 +480,7 @@ class TestRubyLexer < Minitest::Test
   end
 
   def test_yylex_lambda_args__20
-    setup_lexer_class Ruby20Parser
+    setup_lexer_class RubyParser::V20
 
     assert_lex("-> (a) { }",
                s(:iter, s(:call, nil, :lambda),
@@ -480,7 +495,7 @@ class TestRubyLexer < Minitest::Test
   end
 
   def test_yylex_lambda_args_opt__20
-    setup_lexer_class Ruby20Parser
+    setup_lexer_class RubyParser::V20
 
     assert_lex("-> (a=nil) { }",
                s(:iter, s(:call, nil, :lambda),
@@ -497,7 +512,7 @@ class TestRubyLexer < Minitest::Test
   end
 
   def test_yylex_lambda_hash__20
-    setup_lexer_class Ruby20Parser
+    setup_lexer_class RubyParser::V20
 
     assert_lex("-> (a={}) { }",
                s(:iter, s(:call, nil, :lambda),
@@ -967,7 +982,17 @@ class TestRubyLexer < Minitest::Test
   end
 
   def test_yylex_float_e_bad_double_e
-    refute_lex "1e2e3"
+    assert_lex3("1e2e3",
+                nil,
+                :tFLOAT, 100, :expr_end,
+                :tIDENTIFIER, "e3", :expr_end)
+  end
+
+  def test_yylex_float_if_modifier
+    assert_lex3("1e2if",
+                nil,
+                :tFLOAT, 100, :expr_end,
+                :kIF_MOD, "if", :expr_beg)
   end
 
   def test_yylex_float_e_bad_trailing_underscore
@@ -1128,6 +1153,47 @@ class TestRubyLexer < Minitest::Test
                 :tNL,             nil,             :expr_beg)
   end
 
+  def test_yylex_heredoc_double_squiggly
+    setup_lexer_class Ruby23Parser
+
+    assert_lex3("a = <<~\"EOF\"\n  blah blah\n  EOF\n\n",
+                nil,
+                :tIDENTIFIER,     "a",           :expr_cmdarg,
+                :tEQL,            "=",           :expr_beg,
+                :tSTRING_BEG,     "\"",          :expr_beg,
+                :tSTRING_CONTENT, "blah blah\n", :expr_beg,
+                :tSTRING_END,     "EOF",         :expr_end,
+                :tNL,             nil,           :expr_beg)
+  end
+
+  # mri handles tabs in a pretty specific way:
+  # https://github.com/ruby/ruby/blob/trunk/parse.y#L5925
+  def test_yylex_heredoc_double_squiggly_with_tab_indentation_remaining
+    setup_lexer_class Ruby23Parser
+
+    assert_lex3("a = <<~\"EOF\"\n  blah blah\n \tblah blah\n  EOF\n\n",
+                nil,
+                :tIDENTIFIER,     "a",                        :expr_cmdarg,
+                :tEQL,            "=",                        :expr_beg,
+                :tSTRING_BEG,     "\"",                       :expr_beg,
+                :tSTRING_CONTENT, "blah blah\n\tblah blah\n", :expr_beg,
+                :tSTRING_END,     "EOF",                      :expr_end,
+                :tNL,             nil,                        :expr_beg)
+  end
+
+  def test_yylex_heredoc_double_squiggly_with_tab_indentation_removed
+    setup_lexer_class Ruby23Parser
+
+    assert_lex3("a = <<~\"EOF\"\n        blah blah\n\t blah blah\n  EOF\n\n",
+                nil,
+                :tIDENTIFIER,     "a",                       :expr_cmdarg,
+                :tEQL,            "=",                       :expr_beg,
+                :tSTRING_BEG,     "\"",                      :expr_beg,
+                :tSTRING_CONTENT, "blah blah\n blah blah\n", :expr_beg,
+                :tSTRING_END,     "EOF",                     :expr_end,
+                :tNL,             nil,                       :expr_beg)
+  end
+
   def test_yylex_heredoc_double_eos
     refute_lex("a = <<\"EOF\"\nblah",
                :tIDENTIFIER, "a",
@@ -1198,6 +1264,19 @@ class TestRubyLexer < Minitest::Test
                 :tNL,             nil,            :expr_beg)
   end
 
+  def test_yylex_heredoc_none_squiggly
+    setup_lexer_class Ruby23Parser
+
+    assert_lex3("a = <<~EOF\n  blah\n  blah\n  EOF\n",
+                nil,
+                :tIDENTIFIER,     "a",            :expr_cmdarg,
+                :tEQL,            "=",            :expr_beg,
+                :tSTRING_BEG,     "\"",           :expr_beg,
+                :tSTRING_CONTENT, "blah\nblah\n", :expr_beg,
+                :tSTRING_END,     "EOF",          :expr_end,
+                :tNL,             nil,            :expr_beg)
+  end
+
   def test_yylex_heredoc_single
     assert_lex3("a = <<'EOF'\n  blah blah\nEOF\n\n",
                 nil,
@@ -1248,6 +1327,19 @@ class TestRubyLexer < Minitest::Test
                 :tNL,             nil,             :expr_beg)
   end
 
+  def test_yylex_heredoc_single_squiggly
+    setup_lexer_class Ruby23Parser
+
+    assert_lex3("a = <<~'EOF'\n  blah blah\n  EOF\n\n",
+                nil,
+                :tIDENTIFIER,     "a",           :expr_cmdarg,
+                :tEQL,            "=",           :expr_beg,
+                :tSTRING_BEG,     "\"",          :expr_beg,
+                :tSTRING_CONTENT, "blah blah\n", :expr_beg,
+                :tSTRING_END,     "EOF",         :expr_end,
+                :tNL,             nil,           :expr_beg)
+  end
+
   def test_yylex_identifier
     assert_lex3("identifier",
                 nil,
@@ -1265,13 +1357,13 @@ class TestRubyLexer < Minitest::Test
   end
 
   def test_yylex_identifier_def__18
-    setup_lexer_class Ruby18Parser
+    setup_lexer_class RubyParser::V18
 
     assert_lex_fname "identifier", :tIDENTIFIER, :expr_end
   end
 
   def test_yylex_identifier_def__1920
-    setup_lexer_class Ruby19Parser
+    setup_lexer_class RubyParser::V19
 
     assert_lex_fname "identifier", :tIDENTIFIER, :expr_endfn
   end
@@ -1308,13 +1400,13 @@ class TestRubyLexer < Minitest::Test
   end
 
   def test_yylex_identifier_equals_def__18
-    setup_lexer_class Ruby18Parser
+    setup_lexer_class RubyParser::V18
 
     assert_lex_fname "identifier=", :tIDENTIFIER, :expr_end
   end
 
   def test_yylex_identifier_equals_def__1920
-    setup_lexer_class Ruby19Parser
+    setup_lexer_class RubyParser::V19
 
     assert_lex_fname "identifier=", :tIDENTIFIER, :expr_endfn
   end
@@ -1409,26 +1501,33 @@ class TestRubyLexer < Minitest::Test
     refute_lex "0d42__24"
   end
 
+  def test_yylex_integer_if_modifier
+    assert_lex3("123if",
+                nil,
+                :tINTEGER, 123, :expr_end,
+                :kIF_MOD, "if", :expr_beg)
+  end
+
   def test_yylex_question_eh_a__18
-    setup_lexer_class Ruby18Parser
+    setup_lexer_class RubyParser::V18
 
     assert_lex3("?a", nil, :tINTEGER, 97, :expr_end)
   end
 
   def test_yylex_question_eh_a__19
-    setup_lexer_class Ruby19Parser
+    setup_lexer_class RubyParser::V19
 
     assert_lex3("?a", nil, :tSTRING, "a", :expr_end)
   end
 
   def test_yylex_question_eh_escape_M_escape_C__18
-    setup_lexer_class Ruby18Parser
+    setup_lexer_class RubyParser::V18
 
     assert_lex3("?\\M-\\C-a", nil, :tINTEGER, 129, :expr_end)
   end
 
   def test_yylex_question_eh_escape_M_escape_C__19
-    setup_lexer_class Ruby19Parser
+    setup_lexer_class RubyParser::V19
 
     assert_lex3("?\\M-\\C-a", nil, :tSTRING, "\M-\C-a", :expr_end)
   end
@@ -1613,14 +1712,14 @@ class TestRubyLexer < Minitest::Test
   end
 
   def test_yylex_open_bracket_exprarg__18
-    setup_lexer_class Ruby18Parser
+    setup_lexer_class RubyParser::V18
     self.lex_state = :expr_arg
 
     assert_lex3(" (", nil, :tLPAREN2, "(", :expr_beg)
   end
 
   def test_yylex_open_bracket_exprarg__19
-    setup_lexer_class Ruby19Parser
+    setup_lexer_class RubyParser::V19
     self.lex_state = :expr_arg
 
     assert_lex3(" (", nil, :tLPAREN_ARG, "(", :expr_beg)
@@ -1785,13 +1884,13 @@ class TestRubyLexer < Minitest::Test
   end
 
   def test_yylex_question__18
-    setup_lexer_class Ruby18Parser
+    setup_lexer_class RubyParser::V18
 
     assert_lex3("?*", nil, :tINTEGER, 42, :expr_end)
   end
 
   def test_yylex_question__19
-    setup_lexer_class Ruby19Parser
+    setup_lexer_class RubyParser::V19
 
     assert_lex3("?*", nil, :tSTRING, "*", :expr_end)
   end
@@ -1810,7 +1909,7 @@ class TestRubyLexer < Minitest::Test
   end
 
   def test_yylex_question_ws_backslashed__18
-    setup_lexer_class Ruby18Parser
+    setup_lexer_class RubyParser::V18
 
     assert_lex3("?\\ ", nil, :tINTEGER, 32, :expr_end)
     assert_lex3("?\\n", nil, :tINTEGER, 10, :expr_end)
@@ -1821,7 +1920,7 @@ class TestRubyLexer < Minitest::Test
   end
 
   def test_yylex_question_ws_backslashed__19
-    setup_lexer_class Ruby19Parser
+    setup_lexer_class RubyParser::V19
 
     assert_lex3("?\\ ", nil, :tSTRING, " ",  :expr_end)
     assert_lex3("?\\n", nil, :tSTRING, "\n", :expr_end)
@@ -1954,6 +2053,14 @@ class TestRubyLexer < Minitest::Test
                 :tREGEXP_END,     "",            :expr_end)
   end
 
+  def test_yylex_regexp_escaped_delim
+    assert_lex3("%r!blah(?\\!blah)!",
+                nil,
+                :tREGEXP_BEG,     "%r\000",       :expr_beg,
+                :tSTRING_CONTENT, "blah(?!blah)", :expr_beg,
+                :tREGEXP_END,     "",             :expr_end)
+  end
+
   def test_yylex_regexp_escape_backslash_terminator_meta1
     assert_lex3("%r{blah\\}blah}",
                 nil,
@@ -2224,7 +2331,7 @@ class TestRubyLexer < Minitest::Test
 
   def test_yylex_string_double_escape_M
     chr = "\341"
-    chr.force_encoding("UTF-8") if RubyLexer::RUBY19
+    chr.force_encoding("UTF-8") if RubyLexer::HAS_ENC
 
     assert_lex3("\"\\M-a\"", nil, :tSTRING, chr, :expr_end)
   end
@@ -2314,7 +2421,7 @@ class TestRubyLexer < Minitest::Test
   end
 
   def test_yylex_string_double_pound_dollar_bad
-    skip if Ruby18Parser === lexer.parser
+    skip if ruby18
 
     assert_lex3('"#$%"', nil,
 
@@ -2496,7 +2603,7 @@ class TestRubyLexer < Minitest::Test
   end
 
   def test_yylex_symbol_zero_byte__18
-    setup_lexer_class Ruby18Parser
+    setup_lexer_class RubyParser::V18
 
     refute_lex(":\"symbol\0\"", :tSYMBEG, ":")
   end
@@ -2631,7 +2738,7 @@ class TestRubyLexer < Minitest::Test
 
   def test_pct_w_backslashes
     ["\t", "\n", "\r", "\v", "\f"].each do |char|
-      next if !RubyLexer::RUBY19 and char == "\v"
+      next if !RubyLexer::HAS_ENC and char == "\v"
 
       assert_lex("%w[foo#{char}bar]",
                  s(:array, s(:str, "foo"), s(:str, "bar")),
@@ -2663,7 +2770,7 @@ class TestRubyLexer < Minitest::Test
   end
 
   def test_yylex_hash_colon_quoted_22
-    setup_lexer_class Ruby22Parser
+    setup_lexer_class RubyParser::V22
 
     assert_lex("{'a':1}",
                s(:hash, s(:lit, :a), s(:lit, 1)),
@@ -2675,7 +2782,7 @@ class TestRubyLexer < Minitest::Test
   end
 
   def test_yylex_hash_colon_quoted_symbol_22
-    setup_lexer_class Ruby22Parser
+    setup_lexer_class RubyParser::V22
 
     assert_lex("{'abc': :b}",
                s(:hash, s(:lit, :abc), s(:lit, :b)),
@@ -2687,7 +2794,7 @@ class TestRubyLexer < Minitest::Test
   end
 
   def test_yylex_hash_colon_double_quoted_symbol_22
-    setup_lexer_class Ruby22Parser
+    setup_lexer_class RubyParser::V22
 
     assert_lex('{"abc": :b}',
                s(:hash, s(:lit, :abc), s(:lit, :b)),
@@ -2698,25 +2805,94 @@ class TestRubyLexer < Minitest::Test
                :tRCURLY, "}", :expr_endarg,   0, 0)
   end
 
-  def test_ruby21_new_numbers
-    skip "Don't have imaginary and rational literal lexing yet"
+  def test_yylex_required_kwarg_no_value_22
+    setup_lexer_class RubyParser::V22
 
-    setup_lexer_class Ruby21Parser
+    assert_lex3("def foo a:, b:\nend",
+                nil,
+                :kDEF, "def", :expr_fname,
+                :tIDENTIFIER, "foo", :expr_endfn,
+                :tLABEL, "a",   :expr_labelarg,
+                :tCOMMA, ",", :expr_beg,
+                :tLABEL, "b",   :expr_labelarg,
+                :tNL, nil, :expr_beg,
+                :kEND, "end", :expr_end)
+  end
 
-    assert_lex3("10r",      nil, :tRATIONAL, "10r", :expr_end)
-    assert_lex3("1.5r",     nil, :tRATIONAL, "1.5r", :expr_end)
+  def test_ruby21_rational_literal
+    setup_lexer_class RubyParser::V21
 
-    assert_lex3("1i",       nil, :tIMAGINARY, "1i", :expr_end)
-    assert_lex3("1+2i",     nil, :tIMAGINARY, "1+2i", :expr_end)
-    assert_lex3("1.2+3.4i", nil, :tIMAGINARY, "1.2+3.4i", :expr_end)
-    assert_lex3("4r+3i",    nil, :tIMAGINARY, "4r+3i", :expr_end)
-    assert_lex3("4r+3ri",   nil, :tIMAGINARY, "4r+3i", :expr_end)
+    assert_lex3("10r",      nil, :tRATIONAL, Rational(10), :expr_end)
+    assert_lex3("0x10r",      nil, :tRATIONAL, Rational(16), :expr_end)
+    assert_lex3("0o10r",      nil, :tRATIONAL, Rational(8), :expr_end)
+    assert_lex3("0or",      nil, :tRATIONAL, Rational(0), :expr_end)
+    assert_lex3("0b10r",      nil, :tRATIONAL, Rational(2), :expr_end)
+    assert_lex3("1.5r",    nil, :tRATIONAL, Rational(15, 10), :expr_end)
+    assert_lex3("15e3r",    nil, :tRATIONAL, Rational(15000), :expr_end)
+    assert_lex3("15e-3r",    nil, :tRATIONAL, Rational(15, 1000), :expr_end)
+    assert_lex3("1.5e3r",    nil, :tRATIONAL, Rational(1500), :expr_end)
+    assert_lex3("1.5e-3r",    nil, :tRATIONAL, Rational(15, 10000), :expr_end)
 
-    assert_lex3("4i+3r",    nil, :tIMAGINARY, "4r+3i", :expr_end) # HACK
-    assert_lex3("1i+2ri",   nil, :tIMAGINARY, "4r+3i", :expr_end) # HACK
+    assert_lex3("-10r", nil,
+                :tUMINUS_NUM, "-", :expr_beg,
+                :tRATIONAL, Rational(10), :expr_end)
+  end
 
-    assert_lex3("1+2ri",    nil, :tIMAGINARY, "1+3ri", :expr_end)
-    refute_lex("1+2ir", :tINTEGER, 1)
+  def test_ruby21_imaginary_literal
+    setup_lexer_class RubyParser::V21
+
+    assert_lex3("1i",       nil, :tIMAGINARY, Complex(0, 1), :expr_end)
+    assert_lex3("0x10i",    nil, :tIMAGINARY, Complex(0, 16), :expr_end)
+    assert_lex3("0o10i",    nil, :tIMAGINARY, Complex(0, 8), :expr_end)
+    assert_lex3("0oi",      nil, :tIMAGINARY, Complex(0, 0), :expr_end)
+    assert_lex3("0b10i",    nil, :tIMAGINARY, Complex(0, 2), :expr_end)
+    assert_lex3("1.5i",    nil, :tIMAGINARY, Complex(0, 1.5), :expr_end)
+    assert_lex3("15e3i",    nil, :tIMAGINARY, Complex(0, 15000), :expr_end)
+    assert_lex3("15e-3i",    nil, :tIMAGINARY, Complex(0, 0.015), :expr_end)
+    assert_lex3("1.5e3i",    nil, :tIMAGINARY, Complex(0, 1500), :expr_end)
+    assert_lex3("1.5e-3i",    nil, :tIMAGINARY, Complex(0, 0.0015), :expr_end)
+
+    assert_lex3("-10i", nil,
+                :tUMINUS_NUM, "-", :expr_beg,
+                :tIMAGINARY, Complex(0, 10), :expr_end)
+  end
+
+  def test_ruby21_rational_imaginary_literal
+    setup_lexer_class RubyParser::V21
+
+    assert_lex3("1ri",       nil, :tIMAGINARY, Complex(0, Rational(1)), :expr_end)
+    assert_lex3("0x10ri",    nil, :tIMAGINARY, Complex(0, Rational(16)), :expr_end)
+    assert_lex3("0o10ri",    nil, :tIMAGINARY, Complex(0, Rational(8)), :expr_end)
+    assert_lex3("0ori",      nil, :tIMAGINARY, Complex(0, Rational(0)), :expr_end)
+    assert_lex3("0b10ri",    nil, :tIMAGINARY, Complex(0, Rational(2)), :expr_end)
+    assert_lex3("1.5ri",    nil, :tIMAGINARY, Complex(0, Rational("1.5")), :expr_end)
+    assert_lex3("15e3ri",    nil, :tIMAGINARY, Complex(0, Rational("15e3")), :expr_end)
+    assert_lex3("15e-3ri",    nil, :tIMAGINARY, Complex(0, Rational("15e-3")), :expr_end)
+    assert_lex3("1.5e3ri",    nil, :tIMAGINARY, Complex(0, Rational("1.5e3")), :expr_end)
+    assert_lex3("1.5e-3ri",    nil, :tIMAGINARY, Complex(0, Rational("1.5e-3")), :expr_end)
+
+    assert_lex3("-10ri", nil,
+                :tUMINUS_NUM, "-", :expr_beg,
+                :tIMAGINARY, Complex(0, Rational(10)), :expr_end)
+  end
+
+  def test_ruby21_imaginary_literal_with_succeeding_keyword
+    skip "Currently does not tokenize correctly"
+
+    setup_lexer_class RubyParser::V21
+
+    assert_lex3("1if", nil,
+                :tINTEGER, 1, :expr_end,
+                :kIF_MOD, "if", :expr_beg)
+    assert_lex3("1rif", nil,
+                :tRATIONAL, Rational(1), :expr_end,
+                :kIF_MOD, "if", :expr_beg)
+    assert_lex3("1.0if", nil,
+                :tFLOAT, 1.0, :expr_end,
+                :kIF_MOD, "if", :expr_beg)
+    assert_lex3("1.0rif", nil,
+                :tRATIONAL, Rational("1.0"), :expr_end,
+                :kIF_MOD, "if", :expr_beg)
 
     flunk
   end
